@@ -9,12 +9,7 @@
     by Emery Berger
     https://emeryberger.com
 
-    usage: # for CPU profiling only
-            scalene test/testme.py
-            # for CPU and memory profiling (Mac OS X)
-            DYLD_INSERT_LIBRARIES=$PWD/libscalene.dylib PYTHONMALLOC=malloc scalene test/testme.py
-            # for CPU and memory profiling (Linux)
-            LD_PRELOAD=$PWD/libscalene.so PYTHONMALLOC=malloc python scalene test/testme.py
+    usage: scalene test/testme.py
 
 """
 
@@ -34,6 +29,11 @@ import traceback
 from collections import defaultdict
 from contextlib import contextmanager
 from functools import lru_cache
+from rich.console import Console
+from rich.syntax import Syntax
+from rich.table import Table
+from rich.text import Text
+from rich import box
 from signal import Handlers, Signals
 from textwrap import dedent
 from types import CodeType, FrameType
@@ -84,6 +84,92 @@ if sys.platform == "win32":
     sys.exit(-1)
 
 
+def parse_args() -> Tuple[argparse.Namespace, List[str]]:
+    usage = dedent(
+        """Scalene: a high-precision CPU and memory profiler.
+        https://github.com/emeryberger/scalene
+        % scalene yourprogram.py
+        """
+    )
+    parser = argparse.ArgumentParser(
+        prog="scalene",
+        description=usage,
+        formatter_class=argparse.RawTextHelpFormatter,
+    )
+    parser.add_argument("prog", type=str, help="program to be profiled")
+    parser.add_argument(
+        "-o",
+        "--outfile",
+        type=str,
+        default=None,
+        help="file to hold profiler output (default: stdout)",
+    )
+    parser.add_argument(
+        "--html",
+        dest="html",
+        action="store_const",
+        const=True,
+        default=False,
+        help="output as HTML (default: text)",
+    )
+    parser.add_argument(
+        "--profile-interval",
+        type=float,
+        default=float("inf"),
+        help="output profiles every so many seconds.",
+    )
+    parser.add_argument(
+        "--wallclock",
+        dest="wallclock",
+        action="store_const",
+        const=True,
+        default=False,
+        help="use wall clock time (default: virtual time)",
+    )
+    parser.add_argument(
+        "--cpu-only",
+        dest="cpuonly",
+        action="store_const",
+        const=True,
+        default=False,
+        help="only profile CPU time (default: profile CPU, memory, and copying)",
+    )
+    # Parse out all Scalene arguments and jam the remaining ones into argv.
+    # https://stackoverflow.com/questions/35733262/is-there-any-way-to-instruct-argparse-python-2-7-to-remove-found-arguments-fro
+    args, left = parser.parse_known_args()
+    return args, left
+
+
+arguments, left = parse_args()
+
+# Load shared objects unless the user specifies "--cpu-only" at the command-line.
+if not arguments.cpuonly:
+    # Load the shared object on Linux.
+    if sys.platform == "linux":
+        if ("LD_PRELOAD" not in os.environ) and ("PYTHONMALLOC" not in os.environ):
+            os.environ["LD_PRELOAD"] = os.path.join(
+                os.path.dirname(__file__), "libscalene.so"
+            )
+            os.environ["PYTHONMALLOC"] = "malloc"
+            args = sys.argv[1:]
+            args = [os.path.basename(sys.executable), "-m", "scalene"] + args
+            result = subprocess.run(args)
+            sys.exit(result.returncode)
+
+    # Similar logic, but for Mac OS X.
+    if sys.platform == "darwin":
+        if ("DYLD_INSERT_LIBRARIES" not in os.environ) and (
+            "PYTHONMALLOC" not in os.environ
+        ):
+            os.environ["DYLD_INSERT_LIBRARIES"] = os.path.join(
+                os.path.dirname(__file__), "libscalene.dylib"
+            )
+            os.environ["PYTHONMALLOC"] = "malloc"
+            args = sys.argv[1:]
+            args = [os.path.basename(sys.executable), "-m", "scalene"] + args
+            result = subprocess.run(args)
+            sys.exit(result.returncode)
+
 Filename = NewType("Filename", str)
 LineNumber = NewType("LineNumber", int)
 ByteCodeIndex = NewType("ByteCodeIndex", int)
@@ -269,6 +355,8 @@ class Scalene:
     __program_path: str = ""
     # where we write profile info
     __output_file: str = ""
+    # if we output HTML or not
+    __html: bool = False
     # how long between outputting stats during execution
     __output_profile_interval: float = float("inf")
     # when we output the next profile
@@ -736,8 +824,8 @@ process."""
             # Add the byte index to the set for this line (if it's not there already).
             Scalene.__bytei_map[fname][line_no].add(bytei)
             curr = before
-            python_frac = 0
-            allocs = 0
+            python_frac = 0.0
+            allocs = 0.0
             # Go through the array again and add each updated current footprint.
             for item in arr:
                 alloc_time, action, count, python_fraction = item
@@ -835,7 +923,7 @@ process."""
     @staticmethod
     def start() -> None:
         """Initiate profiling."""
-        os.chdir(Scalene.__program_path)
+        # os.chdir(Scalene.__program_path)
         Scalene.enable_signals()
         Scalene.__start_time = Scalene.gettime()
 
@@ -844,7 +932,7 @@ process."""
         """Complete profiling."""
         Scalene.disable_signals()
         Scalene.__elapsed_time += Scalene.gettime() - Scalene.__start_time
-        os.chdir(Scalene.__original_path)
+        # os.chdir(Scalene.__original_path)
 
     # from https://stackoverflow.com/questions/9836370/fallback-to-stdout-if-no-file-name-provided
     @staticmethod
@@ -875,14 +963,24 @@ process."""
 
     @staticmethod
     def output_profile_line(
-        fname: Filename, line_no: LineNumber, line: str, out: IO[str]
+        fname: Filename,
+        line_no: LineNumber,
+        line: str,
+        console: Console,
+        tbl: Table,
     ) -> None:
-        """Print exactly one line of the profile to out."""
+        """Print exactly one line of the profile."""
         current_max = Scalene.__max_footprint
         did_sample_memory: bool = (
             Scalene.__total_memory_free_samples + Scalene.__total_memory_malloc_samples
         ) > 0
-        line = line.rstrip()  # Strip newline
+        # Strip newline
+        line = line.rstrip()
+        # Generate syntax highlighted version.
+        if Scalene.__html:
+            syntax_highlighted = Syntax(line, "python", theme="default", line_numbers=False)
+        else:
+            syntax_highlighted = Syntax(line, "python", theme="vim", line_numbers=False)
         # Prepare output values.
         n_cpu_samples_c = Scalene.__cpu_samples_c[fname][line_no]
         # Correct for negative CPU sample counts. This can happen
@@ -951,7 +1049,7 @@ process."""
             "" if not n_usage_fraction else "%3.0f%%" % (100 * n_usage_fraction)
         )
         n_python_fraction_str: str = (
-            "" if not n_python_fraction else "%3.0f%%" % (100 * n_python_fraction)
+            "" if not n_python_fraction else "%5.0f%%" % (100 * n_python_fraction)
         )
         n_copy_b = Scalene.__memcpy_samples[fname][line_no]
         n_copy_mb_s = n_copy_b / (1024 * 1024 * Scalene.__elapsed_time)
@@ -969,45 +1067,53 @@ process."""
                 )
 
             # Red highlight
+            ncpps : Any = ""
+            ncpcs : Any = ""
+            nufs : Any = ""
             if (
                 n_usage_fraction >= Scalene.__highlight_percentage
                 or (n_cpu_percent_c + n_cpu_percent_python)
                 >= Scalene.__highlight_percentage
             ):
-                print(u"\u001b[31m", end="")
+                ncpps = Text.assemble((n_cpu_percent_python_str, "bold red"))
+                ncpcs = Text.assemble((n_cpu_percent_c_str, "bold red"))
+                nufs = Text.assemble((spark_str + n_usage_fraction_str, "bold red"))
+                # print(u"\u001b[31m", end="")
+            else:
+                ncpps = n_cpu_percent_python_str
+                ncpcs = n_cpu_percent_c_str
+                nufs = spark_str + n_usage_fraction_str
 
-            print(
-                "%6d |%7s |%7s | %5s | %5s | %-9s %-4s |%-6s | %s"
-                % (
-                    line_no,
-                    n_cpu_percent_python_str,
-                    n_cpu_percent_c_str,
-                    n_python_fraction_str,
-                    n_growth_mb_str,
-                    spark_str,
-                    n_usage_fraction_str,
-                    n_copy_mb_s_str,
-                    line,
-                ),
-                file=out,
+            tbl.add_row(
+                str(line_no),
+                ncpps,  # n_cpu_percent_python_str,
+                ncpcs,  # n_cpu_percent_c_str,
+                n_python_fraction_str,
+                n_growth_mb_str,
+                nufs,  # spark_str + n_usage_fraction_str,
+                n_copy_mb_s_str,
+                syntax_highlighted,
             )
+
         else:
 
             # Red highlight
             if (
                 n_cpu_percent_c + n_cpu_percent_python
             ) >= Scalene.__highlight_percentage:
-                print(u"\u001b[31m", end="")
+                ncpps = Text.assemble((n_cpu_percent_python_str, "bold red"))
+                ncpcs = Text.assemble((n_cpu_percent_c_str, "bold red"))
+            else:
+                ncpps = n_cpu_percent_python_str
+                ncpcs = n_cpu_percent_c_str
 
-            print(
-                "%6d |%7s |%7s | %s"
-                % (line_no, n_cpu_percent_python_str, n_cpu_percent_c_str, line),
-                file=out,
+            tbl.add_row(
+                str(line_no),
+                ncpps,  # n_cpu_percent_python_str,
+                ncpcs,  # n_cpu_percent_c_str,
+                syntax_highlighted,
             )
 
-        # Reset color
-        print(u"\u001b[0m", end="")
-
     @staticmethod
     def output_profiles() -> bool:
         """Write the profile out (currently to stdout)."""
@@ -1036,78 +1142,78 @@ process."""
         did_sample_memory: bool = (
             Scalene.__total_memory_free_samples + Scalene.__total_memory_malloc_samples
         ) > 0
-        with Scalene.file_or_stdout(Scalene.__output_file) as out:
+        title = Text()
+        # title.append("scalene\n", style="bold")
+        mem_usage_line = ""
+        if did_sample_memory:
+            samples = Scalene.__memory_footprint_samples
+            if len(samples.get()) > 0:
+                # Output a sparkline as a summary of memory usage over time.
+                _, _, spark_str = Scalene.generate_sparkline(
+                    samples.get()[0 : samples.len()], 0, current_max
+                )
+                mem_usage_line = Text.assemble("Memory usage: ", ((spark_str, "blue")),
+                                               (" (max: %6.2fMB)\n" % current_max))
+                title.append(mem_usage_line)
+
+        null = open("/dev/null", "w")
+        console = Console(width=132, record=True, force_terminal=True, file=null)
+        for fname in sorted(all_instrumented_files):
+
+            fname = Filename(fname)
+            this_cpu_samples = sum(Scalene.__cpu_samples_c[fname].values()) + sum(
+                Scalene.__cpu_samples_python[fname].values()
+            )
+
+            try:
+                percent_cpu_time = (
+                    100 * this_cpu_samples / Scalene.__total_cpu_samples
+                )
+            except ZeroDivisionError:
+                percent_cpu_time = 0
+
+            # Ignore files responsible for less than 1% of execution time,
+            # as long as we aren't profiling memory consumption.
+            if not did_sample_memory and percent_cpu_time < 1:
+                continue
+
+            # Print header.
+            new_title = mem_usage_line + ("%s: %% of CPU time = %6.2f%% out of %6.2fs."
+                                 % (fname, percent_cpu_time, Scalene.__elapsed_time))
+            # Only display total memory usage once.
+            mem_usage_line = ""
+
+            tbl = Table(box=box.MINIMAL_HEAVY_HEAD, title=new_title, collapse_padding=True)
+
+            tbl.add_column("Line", justify="right", no_wrap=True)
+            tbl.add_column("CPU %\nPython", no_wrap=True)
+            tbl.add_column("CPU %\nnative", no_wrap=True)
             if did_sample_memory:
-                samples = Scalene.__memory_footprint_samples
-                if len(samples.get()) > 0:
-                    # Output a sparkline as a summary of memory usage over time.
-                    _, _, spark_str = Scalene.generate_sparkline(
-                        samples.get()[0 : samples.len()], 0, current_max
+                tbl.add_column("Mem %\nPython", no_wrap=True)
+                tbl.add_column("Net\n(MB)", no_wrap=True)
+                tbl.add_column("Memory usage\nover time / %", no_wrap=True)
+                tbl.add_column("Copy\n(MB/s)", no_wrap=True)
+            tbl.add_column("\n" + fname)
+
+            with open(fname, "r") as source_file:
+                for line_no, line in enumerate(source_file, 1):
+                    Scalene.output_profile_line(
+                        fname, LineNumber(line_no), line, console, tbl
                     )
-                    print(
-                        "Memory usage: " + spark_str + " (max: %6.2fMB)" % current_max,
-                        file=out,
-                    )
-
-            for fname in sorted(all_instrumented_files):
-
-                fname = Filename(fname)
-                this_cpu_samples = sum(Scalene.__cpu_samples_c[fname].values()) + sum(
-                    Scalene.__cpu_samples_python[fname].values()
-                )
-
-                try:
-                    percent_cpu_time = (
-                        100 * this_cpu_samples / Scalene.__total_cpu_samples
-                    )
-                except ZeroDivisionError:
-                    percent_cpu_time = 0
-
-                # Ignore files responsible for less than 1% of execution time,
-                # as long as we aren't profiling memory consumption.
-                if not did_sample_memory and percent_cpu_time < 1:
-                    continue
-
-                # Print header.
-                print(
-                    "%s: %% of CPU time = %6.2f%% out of %6.2fs."
-                    % (fname, percent_cpu_time, Scalene.__elapsed_time),
-                    file=out,
-                )
-
-                print(
-                    "       |%7s |%7s | %s %s %s %s"
-                    % (
-                        "CPU %",
-                        "CPU %",
-                        "Mem % |" if did_sample_memory else "",
-                        " Net  |" if did_sample_memory else "",
-                        "Memory usage   |" if did_sample_memory else "",
-                        "Copy  |" if did_sample_memory else "",
-                    ),
-                    file=out,
-                )
-                print(
-                    "  Line |%7s |%7s | %s %s %s %s [%s]"
-                    % (
-                        "Python",
-                        "native",
-                        "Python|" if did_sample_memory else "",
-                        " (MB) |" if did_sample_memory else "",
-                        "over time /  % |" if did_sample_memory else "",
-                        "(MB/s)|" if did_sample_memory else "",
-                        fname,
-                    ),
-                    file=out,
-                )
-                print("-" * 119, file=out)
-
-                with open(fname, "r") as source_file:
-                    for line_no, line in enumerate(source_file, 1):
-                        Scalene.output_profile_line(
-                            fname, LineNumber(line_no), line, out
-                        )
-                    print("", file=out)
+            console.print(tbl)
+            
+        if Scalene.__html:
+            # Write HTML file.
+            if not Scalene.__output_file:
+                Scalene.__output_file = "/dev/stdout"
+            console.save_html(Scalene.__output_file, clear=False)
+        else:
+            if not Scalene.__output_file:
+                # No output file specified: write to stdout.
+                console.save_text("/dev/stdout", styles=True)
+            else:
+                # Don't output styles to text file.
+                console.save_text(Scalene.__output_file, styles=False, clear=False)
         return True
 
     @staticmethod
@@ -1138,54 +1244,15 @@ process."""
     @staticmethod
     def main() -> None:
         """Invokes the profiler from the command-line."""
-        usage = dedent(
-            """Scalene: a high-precision CPU and memory profiler.
-            https://github.com/emeryberger/Scalene
-
-                for CPU profiling only:
-            % python -m scalene yourprogram.py
-                for CPU and memory profiling (Mac OS X):
-            % DYLD_INSERT_LIBRARIES=$PWD/libscalene.dylib PYTHONMALLOC=malloc python -m scalene yourprogram.py
-                for CPU and memory profiling (Linux):
-            % LD_PRELOAD=$PWD/libscalene.so PYTHONMALLOC=malloc python -m scalene yourprogram.py
-            """
-        )
-        parser = argparse.ArgumentParser(
-            prog="scalene",
-            description=usage,
-            formatter_class=argparse.RawTextHelpFormatter,
-        )
-        parser.add_argument("prog", type=str, help="program to be profiled")
-        parser.add_argument(
-            "-o",
-            "--outfile",
-            type=str,
-            default=None,
-            help="file to hold profiler output (default: stdout)",
-        )
-        parser.add_argument(
-            "--profile-interval",
-            type=float,
-            default=float("inf"),
-            help="output profiles every so many seconds.",
-        )
-        parser.add_argument(
-            "--wallclock",
-            dest="wallclock",
-            action="store_const",
-            const=True,
-            default=False,
-            help="use wall clock time (default: virtual time)",
-        )
-        # Parse out all Scalene arguments and jam the remaining ones into argv.
-        # https://stackoverflow.com/questions/35733262/is-there-any-way-to-instruct-argparse-python-2-7-to-remove-found-arguments-fro
-        args, left = parser.parse_known_args()
+        args, left = parse_args()  # We currently do this twice, but who cares.
         sys.argv = sys.argv[:1] + left
         Scalene.set_timer_signal(args.wallclock)
         Scalene.__output_profile_interval = args.profile_interval
         Scalene.__next_output_time = (
             Scalene.gettime() + Scalene.__output_profile_interval
         )
+        Scalene.__html = args.html
+        Scalene.__output_file = args.outfile
         try:
             with open(args.prog, "rb") as prog_being_profiled:
                 Scalene.__original_path = os.getcwd()
@@ -1203,7 +1270,6 @@ process."""
                 # Splice in the name of the file being executed instead of the profiler.
                 the_globals["__file__"] = os.path.basename(args.prog)
                 # Start the profiler.
-                Scalene.__output_file = args.outfile
                 fullname = os.path.join(program_path, os.path.basename(args.prog))
                 profiler = Scalene(Filename(fullname))
                 try:
